#include <jni.h>
#include <pthread.h>

#include <android/log.h>
#ifdef DEBUG
#define LOGV(...) __android_log_print(ANDROID_LOG_VERBOSE, "nativecode", __VA_ARGS__)
#define LOGD(...) __android_log_print(ANDROID_LOG_DEBUG  , "nativecode", __VA_ARGS__)
#define LOGI(...) __android_log_print(ANDROID_LOG_INFO   , "nativecode", __VA_ARGS__)
#define LOGW(...) __android_log_print(ANDROID_LOG_WARN   , "nativecode", __VA_ARGS__)
#define LOGE(...) __android_log_print(ANDROID_LOG_ERROR  , "nativecode", __VA_ARGS__)
#else
#define LOGV(...)
#define LOGD(...)
#define LOGI(...)
#define LOGW(...)
#define LOGE(...)
#endif /* DEBUG */

/* Dimensions of the whole buffer, regardless of slicing */
jint width, height;

/* Thread-specific data */
struct threadData {
	pthread_t id;
	jint sliceid;
	jint *pixels;
	jint offset;
	jint height;
	jint color;
};

/*
 * Adapted from well documented example at http://warp.povusers.org/Mandelbrot/
 */
void *threadFunc(void *data)
{
	struct threadData *args = data;
	double MinRe = -2.0;
	double MaxRe = 1.0;
	double MinIm = -1.2;
	double MaxIm = MinIm + (MaxRe - MinRe) * height / width;
	double Re_factor = (MaxRe - MinRe) / (width - 1);
	double Im_factor = (MaxIm - MinIm) / (height - 1);
	unsigned int x, y, isInside, n;

	LOGI("Slice %d started (offset: %d) (slice height: %d)", args->sliceid, args->offset, args->height);

	for(y = args->offset; y < (args->offset + args->height); y++) {
	    double c_im = MaxIm - y * Im_factor;
	    for(x = 0; x < width; x++) {
	        double c_re = MinRe + x*Re_factor;
	        double Z_re = c_re;
	        double Z_im = c_im;

	        isInside = 1;

	        /* I think 32 iterations give the prettiest picture :) */
	        for(n = 0; n < 32; n++) {
	            double Z_re2 = Z_re*Z_re;
	            double Z_im2 = Z_im*Z_im;

	            if(Z_re2 + Z_im2 > 4) {
	                isInside = 0;
	                break;
	            }

	            Z_im = 2*Z_re*Z_im + c_im;
	            Z_re = Z_re2 - Z_im2 + c_re;
	        }

	        /* Scale 0-32 => 0-256 */
        	n = n * 8;

	        if(isInside)
	        	/* Make members of the set black */
	        	args->pixels[(y * width) + x] = 0xff000000; //0xffff2222;
	        else {
	        	/* Make non-members pretty colours */
	        	unsigned char rval, gval, bval;
	        	unsigned int color;
	        	rval = (n >> 1) < 1 ? 1 : (n >> 1);
	        	gval = (n >> 2) < 1 ? 1 : (n >> 2);
	        	bval = (n) > 255 ? 255 : (n);

	        	/* Alpha | Red | Green | Blue => ARGB_8888 for Android Bitmap */
	        	color = (0xff << 24) | (rval << 16) | (gval << 8) | (bval << 0);
	        	args->pixels[(y * width) + x] = color;
	        }
	    }
	}

	LOGI("Slice %d complete (pixels: %d)", args->sliceid, y * x);

	return NULL;
}


/*
 * Class:     com_arm_android_example_ndkmcpkg_ndkmcact
 * Method:    getNextFrame
 * Signature: ([I)I
 */
JNIEXPORT jint JNICALL Java_com_arm_android_example_ndkmcpkg_ndkmcact_getNextFrame
  (JNIEnv *env, jobject obj, jintArray pixels, jint Jwidth, jint Jheight, jint numthreads)
{
	int i, x, y;
	int sliceheight, offset;
	struct threadData *threads;

	width  = Jwidth;
	height = Jheight;

	/* Allocate space for the thread-specific structs */
	threads = (struct threadData *)malloc(numthreads * sizeof(struct threadData));

	/* Turn our jintArray into a C accessible jint[] and add a reference count to it,
	 * lest Dalvik decides to move it around (or reclaim the space!) :)
	 */
	jint* pixelArr = (*env)->GetIntArrayElements(env, pixels, 0);

	/*
	 * The parallellization in this function is created by splitting the buffer up
	 * into horizontal slices of the entire buffer, and then rendering each slice
	 * on a separate thread.
	 */
	sliceheight = (height / numthreads);

	/*
	 * If height not evenly divisible into the number of slices we want, increase
	 * height and handle a smaller final slice lower down.
	 */
	if ((height % numthreads) != 0)
		sliceheight++;

	LOGI("heightxwidth: %dx%d", height, width);
	LOGI("numthreads: %d", numthreads);
	LOGI("sliceheight: %d", sliceheight);

	/* Set up data structures and spawn threads */
	for (i=0, offset=0; i<numthreads ; i++) {
		struct threadData *curthread = &threads[i];
		curthread->sliceid = i;       /* Only used for logging really */
		curthread->pixels = pixelArr;
		curthread->offset = offset;   /* This slice's offset into total buffer */

		LOGI("offset: %d", curthread->offset);

		if (i != (numthreads - 1)) {
			/* If not last (or only) thread, just set height and update total offset */
			curthread->height = sliceheight;

			offset += sliceheight;
		} else {
			/* If last thread - ensure we don't overrun end of buffer */
			curthread->height = height - ((numthreads - 1) * sliceheight);
		}

		LOGI("height: %d", curthread->height);

		/* Launch a thread for rendering the current slice */
		pthread_create(&curthread->id, NULL, threadFunc, (void *)curthread);
	}

	/* Wait for all threads to complete before returning */
	for (i=0 ; i<numthreads ; i++)
		pthread_join(threads[i].id, NULL);

	/* Toto, I have a feeling we're not in Java anymore... */
	free(threads);

	/*
	 * Let Dalvik know we're done with mucking about with "pixels"
	 */
	(*env)->ReleaseIntArrayElements(env, pixels, pixelArr, 0);
	return 0;
}
